unit Globals;

{

   COPYRIGHT 2013 .. 2015 Blue Hell / Jan Punter

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License version 2 as
  published by the Free Software Foundation;

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  For all listed email addresses :

    _dot. to be substituted by a dot      '.'
    2@t2  to be substituted by an at sign '@'


  Blue Hell is a trade mark owned by

    Jan Punter
    Oogstplein 6
    7545 HP Enschede
    the Netherlands
    http://www.iaf.nl/Users/BlueHell/
    http://bluehell.electro-music.com/
    j_dot.punter2@t2iaf_dot.nl

  All rights attributed to Blue Hell are owned by Jan Punter.
}



// General purpose stuff


interface

uses

  System.SysUtils, WinApi.Windows, System.Classes, WinApi.MMSystem, Vcl.Forms, System.IniFiles, System.Win.Registry,
  Winapi.ShellAPI;


const

  BaseUrl = 'http://bluehell.electro-music.com/modules/';


  function  ApplicationPath    : string;
  function  DocsPath           : string;
  function  ApplicationFileName: string;
  function  IniFileName        : string;
  function  LogFileName        : string;
  function  ForthFileName      : string;
  function  ScaleFileName      : string;
  function  SetPriorityLevel( aLevel: Integer): Boolean;
  procedure StringTofile( const S, aFileName: string);
  function  CreateFileNames( const aFolder, aPattern: string): TStrings;
  procedure ClearDebugTiming;
  function  GetDebugTimeStr: string;
  function  BrowseURL( const anURL: string) : boolean;
  function  CleanFileName( const aFileName : string): string;
  function  BytesFromByteArray( const aSrc: array of Byte; aCount: Integer): TBytes;
  procedure ByteArrayFromBytes( const aSrc: TBytes; out aDst: array of Byte; aCount: Integer);


type


  TRecentStrings = class( TStringList)
  private
    FMaxDepth : Integer;
  private
    procedure   SetMaxDepth( aValue: Integer);
  public
    constructor Create( aMaxDepth : Integer);
    procedure   SetMostRecent( const aName: string);
    procedure   SaveToIni  ( anIniFile: TCustomIniFile; const aSection: string);
    procedure   LoadFromIni( anIniFile: TCustomIniFile; const aSection: string);
  public
    property    MaxDepth: Integer read FMaxDepth write SetMaxDepth;
  end;


  function  GetCompanyName      : string;
  function  GetFileDescription  : string;
  function  GetFileVersion      : string;
  function  GetInternalName     : string;
  function  GetLegalCopyRight   : string;
  function  GetLegalTradeMarks  : string;
  function  GetOriginalFileName : string;
  function  GetProductName      : string;
  function  GetProductVersion   : string;
  function  GetComments         : string;


implementation

// Our name and path stuff

var

  strApplicationPath     : string;
  strDocsPath            : string;
  strApplicationFileName : string;
  strIniFileName         : string;
  strLogFileName         : string;
  strForthFileName       : string;
  strScaleFileName       : string;


  function ApplicationPath: string;
  begin
    Result := strApplicationPath;
  end;

  function DocsPath: string;
  begin
    Result := strDocsPath;
  end;

  function ApplicationFileName: string;
  begin
    Result := strApplicationFileName;
  end;

  function IniFileName: string;
  begin
    Result := strIniFileName;
  end;

  function LogFileName: string;
  begin
    Result := strLogFileName;
  end;

  function ForthFileName: string;
  begin
    Result := strForthFileName;
  end;

  function ScaleFileName: string;
  begin
    Result := strScaleFileName;
  end;

// Date time stuff


  function  SetPriorityLevel( aLevel: Integer): Boolean;
  begin
    Result := False;
    case aLevel of
      -1 : Result := Setpriorityclass( GetCurrentProcess(), IDLE_PRIORITY_CLASS    );
       0 : Result := Setpriorityclass( GetCurrentProcess(), NORMAL_PRIORITY_CLASS  );
       1 : Result := Setpriorityclass( GetCurrentProcess(), HIGH_PRIORITY_CLASS    );
       2 : Result := Setpriorityclass( GetCurrentProcess(), REALTIME_PRIORITY_CLASS);
    end;
  end;

  procedure StringTofile( const S, aFileName: string);
  var
    aFile: TextFile;
  begin
    AssignFile( aFile, afileName);
    try
      Rewrite( aFile);
      Write( aFile, S);
    finally
      CloseFile( aFile);
    end;
  end;

  function  CreateFileNames( const aFolder, aPattern: string): TStrings;
  var
    sr  : TSearchRec;
    res : Integer;
  begin
    Result := nil;
    res    := FindFirst( IncludeTrailingPathDelimiter( aFolder) + aPattern, faAnyFile - faDirectory, sr);
    try
      if res = 0
      then begin
        Result := TStringList.Create;
        try
          while res = 0
          do begin
            Result.Add( IncludeTrailingPathDelimiter( aFolder) + sr.FindData.cFileName);
            res := FindNext( sr);
          end;
        except
          FreeAndNil( Result);
          raise;
        end;
      end;
    finally
      System.SysUtils.FindClose( sr);
    end;
  end;

  function  BytesFromByteArray( const aSrc: array of Byte; aCount: Integer): TBytes;
  begin
    if aCount > Length( aSrc)
    then aCount := Length( aSrc);
    SetLength( Result, aCount);
    Move( aSrc[ Low( aSrc)], Result[ 0], aCount); // Move checks size to be > 0
  end;

  procedure ByteArrayFromBytes( const aSrc: TBytes; out aDst: array of Byte; aCount: Integer);
  begin
    if aCount > Length( aDst)
    then aCount := Length( aDst);
    if aCount > Length( aSrc)
    then aCount := Length( aSrc);
    Move( aSrc[ 0], aDst[ Low( aDst)], aCount);   // Move checks size to be > 0
  end;



// ///////////// Debug timestamp support ///////////

var

  DebugName : string;
  DebugTime : LongWord;


  procedure ClearDebugTiming;
  begin
    DebugTime := TimeGetTime;
  end;

  function FormatDebugTime( aTime: Integer): string;
  var
    min, sec, ms : Integer;
  begin
    ms    := aTime mod 1000;
    aTime := aTime div 1000;
    sec   := aTime mod   60;
    aTime := aTime div   60;
    min   := aTime;
    Result := Format( '%.2d:%.2d %.3d', [ min, sec, ms]);
  end;

  function  GetDebugtime: Integer;
  begin
    Result := TimeGetTime - DebugTime;
  end;

  function  GetDebugTimeStr: string;
  begin
    Result := FormatDebugTime( GetDebugTime);
  end;

  // ///////////////////////////////////////////////


  function  BrowseURL( const anURL: string) : boolean;
  var
    Browser: string;
  begin
     Result := False;
     Browser := '';
     with TRegistry.Create
     do begin
       try
         RootKey := HKEY_CLASSES_ROOT;
         Access  := KEY_QUERY_VALUE;
         if OpenKey( '\http\shell\open\command', False)
         then Browser := ReadString( '');
         CloseKey;
       finally
         DisposeOf;
       end;
     end;
     if Browser <> ''
     then begin
       Browser := Copy( Browser, Pos( '"', Browser) + 1, Length( Browser));
       Browser := Copy( Browser, 1, Pos( '"', Browser) - 1);
       Result  := ShellExecute( 0, 'open', PChar( Browser), PChar( anURL), nil, SW_SHOW) >= 32;
     end;
     if not Result
     then Result := ShellExecute( 0, 'open', PChar( anURL), nil, nil, SW_SHOW) >= 32
  end;


const

  InfoNum  = 10;

var

  FileInfo : Array[ 1 .. InfoNum] Of String = (
    'Unknown',
    'Unknown',
    'Unknown',
    'Unknown',
    'Unknown',
    'Unknown',
    'Unknown',
    'Unknown',
    'Unknown',
    'Unknown'
  );


  procedure InitModule;
  const
    InfoStr : Array[ 1 .. InfoNum] Of String = (
      'CompanyName',
      'FileDescription',
      'FileVersion',
      'InternalName',
      'LegalCopyRight',
      'LegalTradeMarks',
      'OriginalFileName',
      'ProductName',
      'ProductVersion',
      'Comments'
    );
  Var
    S      : String;
    n      : Cardinal;
    Len    : Cardinal;
    Buf    : PChar;
    aValue : PChar;
    i      : Integer;
    p      : Integer;
  begin
    S := Application.ExeName;
    strApplicationFileName := ExtractFileName( S);
    strApplicationPath     := ExtractFilePath( S);
    strDocspath            := strApplicationPath + 'docs';
    p := Pos( '.', strApplicationFileName);
    if p = 0
    then begin
      strIniFileName   := strApplicationPath + strApplicationFileName + '.ini';
      strLogFileName   := strApplicationPath + strApplicationFileName + '.log';
      strForthFileName := strApplicationPath + strApplicationFileName + '.4th';
      strScaleFileName := strApplicationPath + strApplicationFileName + '.scale';
    end
    else begin
      strIniFileName   := strApplicationPath + Copy( strApplicationFileName, 0, p - 1) + '.ini';
      strLogFileName   := strApplicationPath + Copy( strApplicationFileName, 0, p - 1) + '.log';
      strForthFileName := strApplicationPath + Copy( strApplicationFileName, 0, p - 1) + '.4th';
      strScaleFileName := strApplicationPath + Copy( strApplicationFileName, 0, p - 1) + '.scale';
    end;

    DebugName := strApplicationPath + 'debug.txt';

    S := Application.ExeName;
    n := GetFileVersionInfoSize( PChar( S), n);
    if n > 0
    then begin
      Buf := AllocMem( n);
      try
        GetFileVersionInfo( PChar( S), 0, n, Buf);
        for i := 1 to InfoNum
        do begin
          if
            VerQueryValue(
              Buf,
              PChar( '\StringFileInfo\040904E4\' + InfoStr[ i]),
              Pointer( aValue),
              Len
            )
          then FileInfo[ i] := aValue;
        end;
      finally
        FreeMem( Buf, n);
      end;
    end;
    Randomize;
  end;


// System stuff

  function  CleanFileName( const aFileName : string): string;
  var
    i : Integer;
  begin
    Result := ExtractFileName( aFileName);
    i := Pos( '.', Result);
    if i > 0
    then Result := Copy( Result, 1, i - 1);
  end;


{
  TRecentStrings = class( TStringList)
  private
    FMaxDepth : Integer;
  public
    property    MaxDepth: Integer read FMaxDepth write SetMaxDepth;
  private
}

    procedure   TRecentStrings.SetMaxDepth( aValue: Integer);
    begin
      if aValue <> FMaxDepth
      then begin
        FMaxDepth := aValue;
        while Count > MaxDepth
        do Delete( Count - 1);
      end;
    end;

//  public

    constructor TRecentStrings.Create( aMaxDepth : Integer);
    begin
      inherited Create;
      MaxDepth := aMaxDepth;
    end;

    procedure   TRecentStrings.SetMostRecent( const aName: string);
    var
      anIndex : Integer;
    begin
      if aName <> ''
      then begin
        anIndex := indexOf( aName);
        if anIndex >= 0
        then begin
          Exchange( 0, anIndex);
          Strings[ 0] := aName;
        end
        else Insert( 0, aName);
        while Count > MaxDepth
        do Delete( Count - 1);
      end;
    end;

    procedure   TRecentStrings.SaveToIni( anIniFile: TCustomIniFile; const aSection: string);
    var
      i : Integer;
    begin
      with anIniFile
      do begin
        EraseSection( aSection);
        WriteInteger( aSection, 'MaxDepth', MaxDepth);
        WriteInteger( aSection, 'Count'   , Count   );
        for i := 0 to Count - 1
        do WriteString( aSection, Format( 'item%d', [ i]), Strings[ i]);
      end;
    end;

    procedure   TRecentStrings.LoadFromIni( anIniFile: TCustomIniFile; const aSection: string);
    var
      N : Integer;
      i : Integer;
    begin
      with anIniFile
      do begin
        MaxDepth := ReadInteger( aSection, 'MaxDepth', MaxDepth);
        N := ReadInteger( aSection, 'Count', 0);
        for i := 0 to N - 1
        do Add( ReadString( aSection, Format( 'item%d', [ i]), ''));
      end;
    end;


  Function  GetCompanyName      : String;
  Begin
    Result := FileInfo[ 1];
  End;

  Function  GetFileDescription  : String;
  Begin
    Result := FileInfo[ 2];
  End;

  Function  GetFileVersion      : String;
  Begin
    Result := FileInfo[ 3];
  End;

  Function  GetInternalName     : String;
  Begin
    Result := FileInfo[ 4];
  End;

  Function  GetLegalCopyRight   : String;
  Begin
    Result := FileInfo[ 5];
  End;

  Function  GetLegalTradeMarks  : String;
  Begin
    Result := FileInfo[ 6];
  End;

  Function  GetOriginalFileName : String;
  Begin
    Result := FileInfo[ 7];
  End;

  Function  GetProductName      : String;
  Begin
    Result := FileInfo[ 8];
  End;

  Function  GetProductVersion   : String;
  Begin
    Result := FileInfo[ 9];
  End;

  Function  GetComments         : String;
  Begin
    Result := FileInfo[ 10];
  End;


// Initialisation:
//   get version info from executable
//   setup debug stuff
//   etc.


{
var

  aBytes  : TBytes;
  anArray : packed array[ 0 .. 127] of Byte;

  procedure TestArrays;
  var
    i : Integer;
  begin
    SetLength( aBytes, 128);
    for i := 0 to 127
    do aBytes[ i] := i;
    ByteArrayFromBytes( aBytes, anArray, 128);
  end;
}

Initialization

{ TestArrays; }
  InitModule;

Finalization

end.


