{
  Copyright 2000 - blue hell / Jan Punter

  this file is part of the friemels program.
}



unit KnobsWorms;

interface

uses

  Windows, Math, SysUtils, Classes, Graphics, Controls, StdCtrls, ExtCtrls, System.Types;

type


  EKnobsWorm        = class( Exception);
  EKnobsInvalidFile = class( EKnobsWorm    );
  EKnobsBounds      = class( EKnobsWorm    );

  TKnobsWormElt    = Byte;
  TKnobsWormData   = array[ 0 .. MaxInt - 1] of TKnobsWormElt;
  PKnobsWormData   = ^TKnobsWormData;
  TKnobsPercentage = 0 .. 100;

  TKnobsGene = class;

  TKnobsOnGetGeneName = procedure( aSender: TObject; aGene: TKnobsGene; var aName: string) of object;
  TKnobsOnWormChanged = procedure( aSender: TObject; aGene: TKnobsGene)                    of object;

  TKnobsWorm = class( TGraphicControl)
  private
    FGene                 : TKnobsGene;
    FAudible              : Boolean;
    FFrozen               : Boolean;
    FOnAudible            : TNotifyEvent;
    FOnFrozen             : TNotifyEvent;
    FClearance            : Byte;
    FBackColor            : TColor;
    FBorderColorNormal    : TColor;
    FBorderColorAudible   : TColor;
    FBorderColorFrozen    : TColor;
    FBorderColorBoth      : TColor;
    FPenColor             : TColor;
    FTransparent          : Boolean;
    FCrossOverProbability : TKnobsPercentage;
    FMutationProbability  : TKnobsPercentage;
    FMutationRange        : TKnobsWormElt;
    FMorphPercentage      : TKnobsPercentage;
    FOnWormChanged        : TKnobsOnWormChanged;
    FDrawLines            : Boolean;
  private
    procedure   MutePeers;
    procedure   UnActivatePeers;
    function    GetData( anIndex: Integer): TKnobsWormElt;
    function    GetSize              : Integer;
    procedure   SetSize              ( aValue: Integer);
    procedure   SetAudible           ( aValue: Boolean);
    procedure   SetFrozen            ( aValue: Boolean);
    procedure   SetClearance         ( aValue: Byte   );
    procedure   SetBackColor         ( aValue: TColor );
    procedure   SetBorderColorNormal ( aValue: TColor );
    procedure   SetBorderColorAudible( aValue: TColor );
    procedure   SetBorderColorFrozen ( aValue: TColor );
    procedure   SetBorderColorBoth   ( aValue: TColor );
    procedure   SetPenColor          ( aValue: TColor );
    procedure   SetTransparent       ( aValue: Boolean);
    procedure   SetDrawLines         ( aValue: Boolean);
  protected
    procedure   Paint;                                                 override;
    procedure   Do_OnDragDrop ( Sender, Source: TObject; X, Y: Integer);
    procedure   Do_OnDragOver ( Sender, Source: TObject; X, Y: Integer; State: TDragState; var Accept: Boolean);
    procedure   Do_OnEndDrag  ( Sender, Target: TObject; X, Y: Integer);
    procedure   Do_OnMouseDown( Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
    procedure   Changed;
  Public
    constructor Create( anOwner: TComponent);                          override;
    destructor  Destroy;                                               override;
    procedure   FillRandom;
    procedure   Mutate;
    procedure   MateWith( aWorm: TKnobsWorm);
    procedure   MorphTo ( aWorm: TKnobsWorm);
    procedure   CopyFrom( aGene: TKnobsGene);                          overload;
    procedure   CopyFrom( aWorm: TKnobsWorm);                          overload;
    procedure   TriggerChange;
  public
    property    Gene                   : TKnobsGene          read FGene;
    property    Data[ anIndex: Integer]: TKnobsWormElt       read GetData; default;
    property    OnWormChanged          : TKnobsOnWormChanged read FOnWormChanged        write FOnWormChanged;
  published
    property    Align;
    property    Size                   : Integer             read GetSize               write SetSize               default 120;
    property    Audible                : Boolean             read FAudible              write SetAudible            default False;
    property    Frozen                 : Boolean             read FFrozen               write SetFrozen             default False;
    property    Clearance              : Byte                read FClearance            write SetClearance          default 10;
    property    BackColor              : TColor              read FBackColor            write SetBackColor          default clWhite;
    property    BorderColorNormal      : TColor              read FBorderColorNormal    write SetBorderColorNormal  default clGray;
    property    BorderColorAudible     : TColor              read FBorderColorAudible   write SetBorderColorAudible default clYellow;
    property    BorderColorFrozen      : TColor              read FBorderColorFrozen    write SetBorderColorFrozen  default clBlue;
    property    BorderColorBoth        : TColor              read FBorderColorBoth      write SetBorderColorBoth    default clLime;
    property    PenColor               : TColor              read FPenColor             write SetPenColor           default clBlack;
    property    Transparent            : Boolean             read FTransparent          write SetTransparent        default False;
    property    CrossOverProbability   : TKnobsPercentage    read FCrossOverProbability write FCrossOverProbability default 33;
    property    MutationProbability    : TKnobsPercentage    read FMutationProbability  write FMutationProbability  default 50;
    property    MutationRange          : TKnobsWormElt       read FMutationRange        write FMutationRange        default  5;
    property    MorphPercentage        : TKnobsPercentage    read FMorphPercentage      write FMorphPercentage      default 20;
    property    DrawLines              : Boolean             read FDrawLines            write SetDrawLines          default True;
  published
    property    OnAudible              : TNotifyEvent        read FOnAudible            write FOnAudible;
    property    OnFrozen               : TNotifyEvent        read FOnFrozen             write FOnFrozen;
  end;

  TKnobsWorms = array[ 0 .. MaxInt div SizeOf( TKnobsWorm) - 1] of TKnobsWorm;
  PKnobsWorms = ^TKnobsWorms;

  TKnobsWormPanel = class( TCustomPanel)
  private
    FWorms                : TList;
    FRowCount             : Integer;
    FColCount             : Integer;
    FSpacing              : Byte;
    FClearance            : Byte;
    FWormWidth            : Byte;
    FWormHeight           : Byte;
    FWormSize             : Integer;
    FBackColor            : TColor;
    FBorderColorNormal    : TColor;
    FBorderColorAudible   : TColor;
    FBorderColorFrozen    : TColor;
    FBorderColorBoth      : TColor;
    FPenColor             : TColor;
    FTransparent          : Boolean;
    FCrossOverProbability : TKnobsPercentage;
    FMutationProbability  : TKnobsPercentage;
    FMutationRange        : TKnobsWormElt;
    FMorphPercentage      : TKnobsPercentage;
    FDrawLines            : Boolean;
    FAudibleWorm          : TKnobsWorm;
    FFrozenWorm           : TKnobsWorm;
    FAudible              : Integer;
    FFrozen               : Integer;
    FOnAudible            : TNotifyEvent;
    FOnFrozen             : TNotifyEvent;
    FOnWormChanged        : TKnobsOnWormChanged;
  private
    procedure   UpdateWorms;
    procedure   PlaceWorms;
    procedure   UpdateCount;
    procedure   AddWorm;
    procedure   DelWorm;
    procedure   DoAudible( aSender: Tobject);
    procedure   DoFrozen ( aSender: TObject);
    procedure   DoResize ( aSender: TObject);
    procedure   DoWormChanged( aSender: TObject; aGene: TKnobsGene);
    function    GetWorm( anIndex: Integer): TKnobsWorm;
    function    GetData( wIndex, dIndex: Integer): TKnobsWormElt;
    function    GetWormCount: Integer;
    procedure   SetRowCount            ( aValue : Integer    );
    procedure   SetColCount            ( aValue : Integer    );
    procedure   SetAudible             ( aValue : Integer    );
    procedure   SetFrozen              ( aValue : Integer    );
    procedure   SetSpacing             ( aValue : Byte       );
    procedure   SetClearance           ( aValue : Byte       );
    procedure   SetWormWidth           ( aValue : Byte       );
    procedure   SetWormHeight          ( aValue : Byte       );
    procedure   SetWormSize            ( aValue : Integer    );
    procedure   SetBackColor           ( aValue : TColor     );
    procedure   SetBorderColorNormal   ( aValue : TColor     );
    procedure   SetBorderColorAudible  ( aValue : TColor     );
    procedure   SetBorderColorFrozen   ( aValue : TColor     );
    procedure   SetBorderColorBoth     ( aValue : TColor     );
    procedure   SetPenColor            ( aValue : TColor     );
    procedure   SetTransparent         ( aValue : Boolean    );
    procedure   SetCrossOverProbability( aValue : TKnobsPercentage);
    procedure   SetMutationProbability ( aValue : TKnobsPercentage);
    procedure   SetMutationRange       ( aValue : TKnobsWormElt   );
    procedure   SetMorphPercentage     ( aValue : TKnobsPercentage);
    procedure   SetDrawLines           ( aValue : Boolean    );
  protected
    procedure   Notification( aComponent: TComponent; anOperation: TOperation); override;
  public
    constructor Create( anOwner: TComponent);                                   override;
    destructor  Destroy;                                                        override;
    procedure   FillRandom;
    procedure   Mutate;
    procedure   Mate;
    procedure   Morph;
    procedure   TriggerChange;
    procedure   SelectHPrev;
    procedure   SelectHNext;
    procedure   SelectVPrev;
    procedure   SelectVNext;
    procedure   ActivateHPrev;
    procedure   ActivateHNext;
    procedure   ActivateVPrev;
    procedure   ActivateVNext;
  public
    property    Worm[ anIndex       : Integer] : TKnobsWorm    read GetWorm;  default;
    property    Data[ wIndex, dIndex: Integer] : TKnobsWormElt read GetData;
    property    AudibleWorm                    : TKnobsWorm    read FAudibleWorm;
    property    FrozenWorm                     : TKnobsWorm    read FFrozenWorm;
    property    WormCount                      : Integer       read GetWormCount;
  published
    property    Align;
    property    BorderStyle;
    property    BorderWidth;
    property    BevelInner     default bvLowered;
    property    BevelOuter     default bvRaised;
    property    BevelWidth;
    property    Caption;
    property    Color;
    property    Enabled;
    property    Height         default 167;
    property    ParentColor;
    property    ParentShowHint;
    property    PopupMenu;
    property    ShowHint;
    property    TabOrder;
    property    TabStop;
    property    Visible;
    property    Width          default 167;
    property    RowCount             : Integer             read FRowCount             write SetRowCount             default   2;
    property    ColCount             : Integer             read FColCount             write SetColCount             default   2;
    property    Audible              : Integer             read FAudible              write SetAudible              default  -1;
    property    Frozen               : Integer             read FFrozen               write SetFrozen               default  -1;
    property    Spacing              : Byte                read FSpacing              write SetSpacing              default   2;
    property    Clearance            : Byte                read FClearance            write SetClearance            default  10;
    property    WormWidth            : Byte                read FWormWidth            write SetWormWidth            default  80;
    property    WormHeight           : Byte                read FWormHeight           write SetWormHeight           default  80;
    property    WormSize             : Integer             read FWormSize             write SetWormSize             default 120;
    property    BackColor            : TColor              read FBackColor            write SetBackColor            default clWhite;
    property    BorderColorNormal    : TColor              read FBorderColorNormal    write SetBorderColorNormal    default clGray;
    property    BorderColorAudible   : TColor              read FBorderColorAudible   write SetBorderColorAudible   default clYellow;
    property    BorderColorFrozen    : TColor              read FBorderColorFrozen    write SetBorderColorFrozen    default clBlue;
    property    BorderColorBoth      : TColor              read FBorderColorBoth      write SetBorderColorBoth      default clLime;
    property    PenColor             : TColor              read FPenColor             write SetPenColor             default clBlack;
    property    Transparent          : Boolean             read FTransparent          write SetTransparent          default False;
    property    CrossOverProbability : TKnobsPercentage    read FCrossOverProbability write SetCrossOverProbability default  33;
    property    MutationProbability  : TKnobsPercentage    read FMutationProbability  write SetMutationProbability  default  50;
    property    MutationRange        : TKnobsWormElt       read FMutationRange        write SetMutationRange        default   5;
    property    MorphPercentage      : TKnobsPercentage    read FMorphPercentage      write SetMorphPercentage      default  20;
    property    DrawLines            : Boolean             read FDrawLines            write SetDrawLines            default True;
  published
    property    OnAudible            : TNotifyEvent        read FOnAudible            write FOnAudible;
    property    OnFrozen             : TNotifyEvent        read FOnFrozen             write FOnFrozen;
    property    OnWormChanged        : TKnobsOnWormChanged read FOnWormChanged        write FOnWormChanged;
  end;

  TKnobsGene = class( TObject)
  private
    FData  : PKnobsWormData;
    FSize  : Integer;
  private
    procedure   SetSize( aValue: Integer);
    function    GetData( anIndex: Integer): TKnobsWormElt;
    procedure   SetData( anIndex: Integer; aValue: TKnobsWormElt);
  public
    constructor Create( aWorm: TKnobsWorm = nil);
    destructor  Destroy;                                               override;
    procedure   Save( var aFile: TextFile);

    property    Size : Integer                         read FSize   write SetSize;
    property    Data[ anIndex: Integer]: TKnobsWormElt read GetData write SetData; default;
  end;

  TKnobsGeneBox = class( TCustomListBox)
  private
    FOnGetGeneName : TKnobsOnGetGeneName;
    FDragged       : TKnobsGene;
  private
    procedure   Do_OnDragDrop ( Sender, Source: TObject; X, Y: Integer);
    procedure   Do_OnDragOver ( Sender, Source: TObject; X, Y: Integer; State: TDragState; Var Accept: Boolean);
    procedure   Do_OnEndDrag  ( Sender, Target: TObject; X, Y: Integer);
    procedure   Do_OnMouseDown( Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
    procedure   AddWorm( aName: string; aWorm: TKnobsWorm);
  public
    constructor Create( anOwner: TComponent);                          override;
    destructor  Destroy;                                               override;
    procedure   Load  ( Const aFileName: String);
    procedure   Append( Const aFileName: String);
    procedure   Save  ( Const aFileName: String);
    procedure   Clear;                                                 override;
  published
    property    Align;
    property    Anchors;
    property    BiDiMode;
    property    BorderStyle;
    property    Color;
    property    Columns;
    property    Constraints;
    property    Ctl3D;
    property    DragCursor;
    property    DragKind;
    property    Enabled;
    property    ExtendedSelect;
    property    Font;
    property    ImeMode;
    property    ImeName;
    property    IntegralHeight;
    property    ItemHeight;
    property    ParentBiDiMode;
    property    ParentColor;
    property    ParentCtl3D;
    property    ParentFont;
    property    ParentShowHint;
    property    PopupMenu;
    property    ShowHint;
    property    Sorted;
    property    Style;
    property    TabOrder;
    property    TabWidth;
    property    Visible;
  published
    property    OnGetGeneName: TKnobsOnGetGeneName read FOnGetGeneName write FOnGetGeneName;
    property    OnClick;
    property    OnContextPopup;
    property    OnDblClick;
    property    OnEnter;
    property    OnExit;
    property    OnKeyDown;
    property    OnKeyPress;
    property    OnKeyUp;
    property    OnMeasureItem;
  end;

  TKnobsBoolArray = array[ 0 .. MaxInt div SizeOf( Boolean) - 1] of Boolean;
  PKnobsBoolArray = ^TKnobsBoolArray;

  TSubstitution = class
  private
    FRule   : TStringList;
    FActive : PKnobsBoolArray;
    FCount  : Integer;
  private
    procedure   CheckBounds( anIndex: Integer);
    function    GetActive( anIndex: Integer): Boolean;
    procedure   SetActive( anIndex: Integer; aValue: Boolean);
    function    GetRule  ( anIndex: Integer): String;
    procedure   SetRule  ( anIndex: Integer; aValue: String);
    procedure   SetCount ( aValue : Integer);
  public
    constructor Create;
    destructor  Destroy;                                               override;
    procedure   CopyFrom( aSubstitution: TSubstitution);
    procedure   Load( const AfileName: string);
    procedure   Save( const AfileName: string);
  public
    property    Active[ anIndex: Integer]: Boolean read GetActive write SetActive;
    property    Rule  [ anIndex: Integer]: String  read GetRule   write SetRule;
    property    Count                    : Integer read FCount    write SetCount;
  end;



implementation





type

  TCharSet = set of AnsiChar;

  function ScanName( const S: string; var p: Integer; aSeparators: TCharSet): string;
  var
    i : Integer;
  begin
    Result := '';

    while (( p <= Length( S)) and CharInSet( S[ p], aSeparators))
    do Inc( p);

    i := p;

    while (( i <= Length( S)) and not CharInSet( S[ i], aSeparators))
    do Inc( i);

    Result := Copy( S, p, i - p);
    p := i;
  end;

  function ScanInt( const S: string; var p: Integer; aSeparators: TCharSet): Integer;
  var
    i : Integer;
    N : string;
  begin
    N := '';

    while (( p <= Length( S)) and CharInSet( S[ p], aSeparators))
    do Inc( p);

    i := p;

    while (( i <= Length( S)) and not CharInSet( S[ i], aSeparators))
    do Inc( i);

    N := Copy( S, p, i - p);
    p := i;
    Result := StrToIntDef( N, 0);
  end;

  function ScanBool( const S: string; var p: Integer; aSeparators: TCharSet): Boolean;
  Var
    i : Integer;
    N : string;
  Begin
    N := '';

    while (( p <= Length( S)) and CharInSet( S[ p], aSeparators))
    do Inc( p);

    i := p;

    while (( i <= Length( S)) and not CharInSet( S[ i], aSeparators))
    do Inc( i);

    N := Copy( S, p, i - p);
    p := i;
    Result := UpperCase( N) = 'TRUE';
  end;


{ ========
  TKnobsWorm = class( TGraphicControl)
  private
    FGene                 : TKnobsGene;
    FAudible              : Boolean;
    FFrozen               : Boolean;
    FOnAudible            : TNotifyEvent;
    FOnFrozen             : TNotifyEvent;
    FClearance            : Byte;
    FBackColor            : TColor;
    FBorderColorNormal    : TColor;
    FBorderColorAudible   : TColor;
    FBorderColorFrozen    : TColor;
    FBorderColorBoth      : TColor;
    FPenColor             : TColor;
    FTransparent          : Boolean;
    FCrossOverProbability : TKnobsPercentage;
    FMutationProbability  : TKnobsPercentage;
    FMutationRange        : TKnobsWormElt;
    FMorphPercentage      : TKnobsPercentage;
    FOnWormChanged        : TKnobsOnWormChanged;
    FDrawLines            : Boolean;
  public
    property    Gene                   : TKnobsGene          read FGene;
    property    Data[ anIndex: Integer]: TKnobsWormElt       read GetData; default;
    property    OnWormChanged          : TKnobsOnWormChanged read FOnWormChanged        write FOnWormChanged;
  published
    property    Align;
    property    Size                   : Integer             read GetSize               write SetSize               default 120;
    property    Audible                : Boolean             read FAudible              write SetAudible            default False;
    property    Frozen                 : Boolean             read FFrozen               write SetFrozen             default False;
    property    Clearance              : Byte                read FClearance            write SetClearance          default 10;
    property    BackColor              : TColor              read FBackColor            write SetBackColor          default clWhite;
    property    BorderColorNormal      : TColor              read FBorderColorNormal    write SetBorderColorNormal  default clGray;
    property    BorderColorAudible     : TColor              read FBorderColorAudible   write SetBorderColorAudible default clYellow;
    property    BorderColorFrozen      : TColor              read FBorderColorFrozen    write SetBorderColorFrozen  default clBlue;
    property    BorderColorBoth        : TColor              read FBorderColorBoth      write SetBorderColorBoth    default clLime;
    property    PenColor               : TColor              read FPenColor             write SetPenColor           default clBlack;
    property    Transparent            : Boolean             read FTransparent          write SetTransparent        default False;
    property    CrossOverProbability   : TKnobsPercentage    read FCrossOverProbability write FCrossOverProbability default 33;
    property    MutationProbability    : TKnobsPercentage    read FMutationProbability  write FMutationProbability  default 50;
    property    MutationRange          : TKnobsWormElt       read FMutationRange        write FMutationRange        default  5;
    property    MorphPercentage        : TKnobsPercentage    read FMorphPercentage      write FMorphPercentage      default 20;
    property    DrawLines              : Boolean             read FDrawLines            write SetDrawLines          default True;
  published
    property    OnAudible              : TNotifyEvent        read FOnAudible            write FOnAudible;
    property    OnFrozen               : TNotifyEvent        read FOnFrozen             write FOnFrozen;
  private
 }

    procedure   TKnobsWorm.MutePeers;
    var
      i : Integer;
    begin
      if Assigned( Parent)
      then begin
        with Parent
        do begin
          for i := 0 to ControlCount - 1
          do begin
            if Controls[ i] is TKnobsWorm
            then TKnobsWorm( Controls[ i]).Audible := False;
          end;
        end;
      end;
    end;

    procedure   TKnobsWorm.UnActivatePeers;
    var
      i : Integer;
    begin
      if Assigned( Parent)
      then begin
        With Parent
        do begin
          for i := 0 to ControlCount - 1
          do begin
            if Controls[ i] is TKnobsWorm
            then TKnobsWorm( Controls[ i]).Frozen:= False;
          end;
        end;
      end;
    end;

    function    TKnobsWorm.GetSize: Integer;
    begin
      Result := FGene.Size;
    end;

    procedure   TKnobsWorm.SetSize ( aValue: Integer);
    begin
      FGene.Size := aValue;
      if not ( csDestroying In ComponentState)
      then FillRandom;
    end;

    function    TKnobsWorm.GetData( anIndex: Integer): TKnobsWormElt;
    begin
      Result := FGene[ anIndex];
    end;

    procedure   TKnobsWorm.SetAudible( aValue: Boolean);
    begin
      if FAudible <> aValue
      then begin
        if aValue
        then MutePeers;

        FAudible := aValue;

        if FAudible and Assigned( FOnAudible)
        then FOnAudible( Self);

        Invalidate;
      end;
    end;

    procedure   TKnobsWorm.SetFrozen( aValue: Boolean);
    begin
      if FFrozen <> aValue
      then Begin
        if aValue
        then UnActivatePeers;

        FFrozen := aValue;

        if FFrozen and Assigned( FOnFrozen)
        then FOnFrozen( Self);

        Invalidate;
      end;
    end;

    Procedure   TKnobsWorm.SetClearance;
    Begin
      If aValue <> FClearance
      Then Begin
        FClearance := aValue;
        Invalidate;
      End;
    End;

    Procedure   TKnobsWorm.SetBackColor( aValue: TColor);
    Begin
      If aValue <> FBackColor
      Then Begin
        FBackColor := aValue;
        Invalidate;
      End;
    End;

    Procedure   TKnobsWorm.SetBorderColorNormal( aValue: TColor);
    Begin
      If aValue <> FBorderColorNormal
      Then Begin
        FBorderColorNormal := aValue;
        Invalidate;
      End;
    End;

    Procedure   TKnobsWorm.SetBorderColorAudible( aValue: TColor);
    Begin
      If aValue <> FBorderColorAudible
      Then Begin
        FBorderColorAudible := aValue;
        Invalidate;
      End;
    End;

    Procedure   TKnobsWorm.SetBorderColorFrozen( aValue: TColor);
    Begin
      If aValue <> FBorderColorFrozen
      Then Begin
        FBorderColorFrozen := aValue;
        Invalidate;
      End;
    End;

    Procedure   TKnobsWorm.SetBorderColorBoth( aValue: TColor);
    Begin
      If aValue <> FBorderColorBoth
      Then Begin
        FBorderColorBoth := aValue;
        Invalidate;
      End;
    End;

    Procedure   TKnobsWorm.SetPenColor( aValue: TColor);
    Begin
      If aValue <> FPenColor
      Then Begin
        FPenColor := aValue;
        Invalidate;
      End;
    End;

    Procedure   TKnobsWorm.SetTransparent( aValue: Boolean);
    Begin
      If aValue <> FTransparent
      Then Begin
        FTransparent := aValue;
        Invalidate;
      End;
    End;

    Procedure   TKnobsWorm.SetDrawLines( aValue: Boolean);
    Begin
      If aValue <> FDrawLines
      Then Begin
        FDrawLines := aValue;
        Invalidate;
      End;
    End;

//   Protected

    procedure   TKnobsWorm.Paint; // Override;
    Type
      TP = Record
        X, Y : Real;
      End;
      TPS = Array[ 0 .. MaxInt Div SizeOf( TP) - 1] Of TP;
      PPS = ^TPS;

    Var
      Pts  : PPS;
      maxx : Real;
      maxy : Real;
      minx : Real;
      miny : Real;

      function P( anX, anY: Real): TP;
      begin
        With Result Do
        Begin
          X := anX;
          y := anY;
        End;
      end;

      procedure ScalePts;
      var
        dx : Real;
        dy : Real;
        i  : Integer;
      begin
        dx := maxx - minx;
        If dx < 1
        Then dx := 1;
        dy := maxy - miny;
        If dy < 1
        Then dy := 1;
        // dx := Max( dx, dy);
        // dy := Max( dx, dy);
        For i := 0 To Size
        Do Begin
          Pts^[ i].X := FClearance + ( Width  - 2 * FClearance) * (( Pts^[ i].X - minx) / dx);
          Pts^[ i].Y := FClearance + ( Height - 2 * FClearance) * (( Pts^[ i].Y - miny) / dy);
        End;
      end;

      procedure CalcPts;
      var
        i  : Integer;
        nx : Real;
        ny : Real;
      begin
        Pts^[ 0] := P( 0, 0);
        maxx     := 0.0;
        maxy     := 0.0;
        minx     := 0.0;
        miny     := 0.0;
        For i := 0 To Size - 1
        Do Begin
          nx := Pts^[ i].X + Sin( 2 * Pi * FGene[ i] / 256);
          ny := Pts^[ i].Y + Cos( 2 * Pi * FGene[ i] / 256);
          If nx > maxx
          Then maxx := nx
          Else If nx < minx
          Then minx := nx;
          If ny > maxy
          Then maxy := ny
          Else If ny < miny
          Then miny := ny;
          Pts^[ i + 1] := P( nx, ny);
        End;
      end;

    var               // Sel   // Act
      Colors : Array[ Boolean, Boolean] Of TColor;
      i      : Integer;
    begin
      Colors[ false, false] := BorderColorNormal;
      Colors[ false, true ] := BorderColorFrozen;
      Colors[ true , false] := BorderColorAudible;
      Colors[ true , true ] := BorderColorBoth;

      GetMem( Pts, ( Size + 1 ) * SizeOf( TP));

      try
        CalcPts;
        ScalePts;

        with Canvas
        do Begin
          Brush.Color := BackColor;

          with Pen
          do Begin
            Color := Colors[ Audible, Frozen];
            Width := 4;
          end;

          Rectangle( 0, 0, Width, Height);

          with Pen
          do Begin
            Color := PenColor;
            Width := 1;
          end;

          if FDrawLines
          then Begin
            MoveTo( Round( Pts^[ 0].X), Round( Pts^[ 0].Y));

            for i := 1 to Size
            do LineTo( Round( Pts^[ i].X), Round( Pts^[ i].Y));
          end
          else Begin
            For i := 1 To Size - 1
            Do begin
              Ellipse(
                Round( Pts^[ i].X) - 2,
                Round( Pts^[ i].Y) - 2,
                Round( Pts^[ i].X) + 2,
                Round( Pts^[ i].Y) + 2
              );
            end;
          end;
        end;
      finally
        FreeMem( Pts, ( Size + 1) * SizeOf( TP));
      end;
    End;

    Procedure   TKnobsWorm.Do_OnDragDrop( Sender, Source: TObject; X, Y: Integer);
    Begin
      Audible := True;
      If ( Source Is TKnobsWorm) And ( Source <> Self)
      Then Begin
        If HasParent And ( Parent Is TKnobsWormPanel)
        Then TKnobsWormPanel( Parent).Mate;
      End
      Else If Source Is TKnobsGeneBox
      Then begin
        With Source as TKnobsGeneBox
        Do CopyFrom( FDragged);
      end;
    End;

    Procedure   TKnobsWorm.Do_OnDragOver( Sender, Source: TObject; X, Y: Integer; State: TDragState; Var Accept: Boolean);
    Begin
      Accept :=
        (( Source Is TKnobsWorm) And ( Source <> Self)) Or
        ( Source Is TKnobsGeneBox)
        ;
    End;

    Procedure   TKnobsWorm.Do_OnEndDrag( Sender, Target: TObject; X, Y: Integer);
    Begin
    End;

    Procedure   TKnobsWorm.Do_OnMouseDown( Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
    Begin
      If Button = mbLeft
      Then Begin
        Audible := True;
        BeginDrag( False, -1);
      End
      Else If Button = mbRight
      Then Frozen := True;
    End;

    Procedure   TKnobsWorm.Changed;
    Var
      aGene : TKnobsGene;
    Begin
      If Audible And Assigned( FOnWormChanged)
      Then Begin
        aGene := TKnobsGene.Create( Self);
        Try
          FOnWormChanged( Self, aGene);
        Finally
          aGene.Free;
        End;
      End;
      Invalidate; // Hmmm ... 't belongs here alright, but why ...
    End;

//  Public

    Constructor TKnobsWorm.Create( anOwner: TComponent); // Override;
    Begin
      Inherited Create( anOwner);
      FGene                := TKnobsGene.Create;
      Size                 := 120;
      Audible              := False;
      Frozen               := False;
      Clearance            := 10;
      BackColor            := clWhite;
      BorderColorNormal    := clGray;
      BorderColorAudible   := clYellow;
      BorderColorFrozen    := clBlue;
      BorderColorBoth      := clLime;
      PenColor             := clBlack;
      Transparent          := False;
      CrossOverProbability := 33;
      MutationProbability  := 50;
      MutationRange        :=  5;
      MorphPercentage      := 20;
      Width                := 80;
      Height               := 80;
      OnDragDrop           := Do_OnDragDrop;
      OnDragOver           := Do_OnDragOver;
      OnEndDrag            := Do_OnEndDrag;
      OnMouseDown          := Do_OnMouseDown;
      DragMode             := dmManual;
      DrawLines            := True;
      FillRandom;
    End;

    Destructor  TKnobsWorm.Destroy; // Override;
    Begin
      Size := 0;
      FGene.Free;
      Inherited Destroy;
    End;

    Procedure   TKnobsWorm.FillRandom;
    Var
      i : Integer;
    Begin
      For i := 0 To Size - 1
      Do FGene[ i] := Random( 256);
      Changed;
    End;

    Procedure   TKnobsWorm.Mutate;
    Var
      i : Integer;
    Begin
      For i := 0 To Size - 1
      Do begin
        If Random( 100) < MutationProbability
        Then FGene[ i] := ( Integer( 256 + FGene[ i]) + MutationRange - Random( 2 * MutationRange + 1) ) Mod 256;
      end;
      Changed;
    End;

    Procedure   TKnobsWorm.MateWith( aWorm: TKnobsWorm);
    Var
      i : Integer;
      S : Boolean;
    Begin
      If Assigned( aWorm)
      Then Begin
        S := Random( 100) < CrossoverProbability;
        For i := 0 To Min( Size, aWorm.Size) - 1
        Do Begin
          If S
          Then FGene[ i] := aWorm.FGene[ i];
          S := S Xor ( Random( 100) < CrossoverProbability);
        End;
        Mutate; // Performs Changed.
      End;
    End;

    Procedure   TKnobsWorm.MorphTo( aWorm: TKnobsWorm);
    Var
      i : Integer;
    Begin
      If Assigned( aWorm)
      Then Begin
        For i := 0 To Min( Size, aWorm.Size) - 1
        Do begin
          FGene[ i] :=
            ((
              (100 - MorphPercentage) *       FGene[ i] +
                     MorphPercentage  * aWorm.FGene[ i]
            ) Div 100 ) Mod 256;
        end;
        Changed;
      End;
    End;

    Procedure   TKnobsWorm.CopyFrom( aGene: TKnobsGene);
    Var
      i : Integer;
    Begin
      If Assigned( aGene)
      Then Begin
        Size := aGene.Size;
        For i := 0 To Size - 1
        Do FGene[ i] := aGene.FData^[ i];
      End;
      Changed;
    End;

    Procedure   TKnobsWorm.CopyFrom( aWorm: TKnobsWorm);
    Begin
      If Assigned( aWorm)
      Then CopyFrom( aWorm.Gene);
    End;

    Procedure   TKnobsWorm.TriggerChange;
    Begin
      Changed;
    End;


{ ========
  TWormPanel = Class( TPanel)
  Private
      FWorms                : TList;
      FRowCount             : Integer;
      FColCount             : Integer;
      FSpacing              : Byte;
      FClearance            : Byte;
      FWormWidth            : Byte;
      FWormHeight           : Byte;
      FWormSize             : Integer;
      FBackColor            : TColor;
      FBorderColorNormal    : TColor;
      FBorderColorAudible   : TColor;
      FBorderColorFrozen    : TColor;
      FBorderColorBoth      : TColor;
      FPenColor             : TColor;
      FTransparent          : Boolean;
      FCrossOverProbability : TPercentage;
      FMutationProbability  : TPercentage;
      FMutationRange        : TWormElt;
      FMorphPercentage      : TPercentage;
      FAudibleWorm          : TWorm;
      FFrozenWorm           : TWorm;
      FAudible              : Integer;
      FFrozen               : Integer;
      FOnAudible            : TNotifyEvent;
      FOnFrozen             : TNotifyEvent;
      FOnWormChanged        : TOnWormChanged;
  Public
    Property    Worm[ anIndex       : Integer] : TWorm    Read GetWorm;  Default;
    Property    Data[ wIndex, dIndex: Integer] : TWormElt Read GetData;
    Property    AudibleWorm                    : TWorm    Read FAudibleWorm;
    Property    FrozenWorm                     : TWorm    Read FFrozenWorm;
    Property    WormCount                      : Integer  Read GetWormCount;
  Published
    Property    Align;
    Property    BorderStyle;
    Property    BorderWidth;
    Property    BevelInner     Default bvLowered;
    Property    BevelOuter     Default bvRaised;
    Property    BevelWidth;
    Property    Caption;
    Property    Color;
    Property    Enabled;
    Property    Height         Default 167;
    Property    ParentColor;
    Property    ParentShowHint;
    Property    PopupMenu;
    Property    ShowHint;
    Property    TabOrder;
    Property    TabStop;
    Property    Visible;
    Property    Width          Default 167;
    Property    RowCount             : Integer      Read FRowCount             Write SetRowCount             Default   2;
    Property    ColCount             : Integer      Read FColCount             Write SetColCount             default   2;
    Property    Audible              : Integer      Read FAudible              Write SetAudible;
    Property    Frozen               : Integer      Read FFrozen               Write SetFrozen;
    Property    Spacing              : Byte         Read FSpacing              Write SetSpacing              Default   2;
    Property    Clearance            : Byte         Read FClearance            Write SetClearance            Default   5;
    Property    WormWidth            : Byte         Read FWormWidth            Write SetWormWidth            Default  80;
    Property    WormHeight           : Byte         Read FWormHeight           Write SetWormHeight           Default  80;
    Property    WormSize             : Integer      Read FWormSize             Write SetWormSize             Default 120;
    Property    BackColor            : TColor       Read FBackColor            Write SetBackColor            Default clWhite;
    Property    BorderColorNormal    : TColor       Read FBorderColorNormal    Write SetBorderColorNormal    Default clGray;
    Property    BorderColorAudible   : TColor       Read FBorderColorAudible   Write SetBorderColorAudible   Default clYellow;
    Property    BorderColorFrozen    : TColor       Read FBorderColorFrozen    Write SetBorderColorFrozen    Default clBlue;
    Property    BorderColorBoth      : TColor       Read FBorderColorBoth      Write SetBorderColorBoth      Default clLime;
    Property    PenColor             : TColor       Read FPenColor             Write SetPenColor             Default clBlack;
    Property    Transparent          : Boolean      Read FTransparent          Write SetTransparent          Default False;
    Property    CrossOverProbability : TPercentage  Read FCrossOverProbability Write SetCrossOverProbability Default  33;
    Property    MutationProbability  : TPercentage  Read FMutationProbability  Write SetMutationProbability  Default  50;
    Property    MutationRange        : TWormElt     Read FMutationRange        Write SetMutationRange        Default   5;
    Property    MorphPercentage      : TPercentage  Read FMorphPercentage      Write SetMorphPercentage      Default  20;
  Published
    Property    OnAudible            : TNotifyEvent   Read FOnAudible     Write FOnAudible;
    Property    OnFrozen             : TNotifyEvent   Read FOnFrozen      Write FOnFrozen;
    Property    OnWormChanged        : TOnWormChanged Read FOnWormChanged Write FOnWormChanged;
  Private
}

    Procedure   TKnobsWormPanel.UpdateWorms;
    Var
      i : Integer;
    Begin
      For i := 0 To WormCount - 1
      Do begin
        With Worm[ i]
        Do Begin
          Clearance            := Self.Clearance;
          BackColor            := Self.BackColor;
          BorderColorNormal    := Self.BorderColorNormal;
          BorderColorAudible   := Self.BorderColorAudible;
          BorderColorFrozen    := Self.BorderColorFrozen;
          BorderColorBoth      := Self.BorderColorBoth;
          PenColor             := Self.PenColor;
          Transparent          := Self.Transparent;
          CrossOverProbability := Self.CrossOverProbability;
          MutationProbability  := Self.MutationProbability;
          MutationRange        := Self.MutationRange;
          MorphPercentage      := Self.MorphPercentage;
          DrawLines            := Self.DrawLines;
          Width                := WormWidth;
          Height               := WormHeight;
        End;
      end;
    End;

    Procedure   TKnobsWormPanel.PlaceWorms;
    Var
      i      : Integer;
      CurCol : Integer;
      CurRow : Integer;
    Begin
      Width  := ColCount * ( WormWidth  + Spacing + 1) + Spacing;
      Height := RowCount * ( WormHeight + Spacing + 1) + Spacing;
      If Not ( csDesigning In ComponentState)
      Then Begin
        CurCol := 0;
        CurRow := 0;
        For i := 0 To RowCount * ColCount - 1
        Do begin
          With Worm[ i]
          Do Begin
            Left := Spacing + ( CurCol * ( Width  + Spacing));
            Top  := Spacing + ( CurRow * ( Height + Spacing));
            Inc( CurCol);
            If CurCol >= ColCount
            Then Begin
              CurCol := 0;
              Inc( CurRow);
            End;
          End;
        end;
      End;
    End;

    Procedure   TKnobsWormPanel.AddWorm;
    Var
      aWorm : TKnobsWorm;
    Begin
      aWorm := TKnobsWorm.Create( Self);
      With aWorm
      Do Begin
        Parent        := Self;
        OnWormChanged := DoWormChanged;
      End;
    End;

    Procedure   TKnobsWormPanel.DelWorm;
    Begin
      Worm[ WormCount - 1].Free;
    End;

    Procedure   TKnobsWormPanel.UpdateCount;
    Var
      aCount : Integer;
    Begin
      If Not ( csDesigning In ComponentState)
      Then Begin
        aCount := FRowCount * FColCount;
        While WormCount < aCount
        Do AddWorm;
        While WormCount > aCount
        Do DelWorm;
      End;
      PlaceWorms;
    End;

    Procedure   TKnobsWormPanel.DoAudible( aSender: Tobject);
    Begin
      If aSender Is TKnobsWorm
      Then Begin
        FAudibleWorm := TKnobsWorm( aSender);
        FAudible     := TKnobsWorm( aSender).Tag;
        If Assigned( FOnAudible)
        Then FOnAudible( Self);
        TriggerChange;
      End;
    End;

    Procedure   TKnobsWormPanel.DoFrozen( aSender: TObject);
    Begin
      If aSender Is TKnobsWorm
      Then Begin
        FFrozenWorm := TKnobsWorm( aSender);
        FFrozen     := TKnobsWorm( aSender).Tag;
        If Assigned( FOnFrozen)
        Then FOnFrozen( Self);
      End;
    End;

    Procedure   TKnobsWormPanel.DoResize( aSender: TObject);
    Begin
      PlaceWorms;
    End;

    Procedure   TKnobsWormPanel.DoWormChanged( aSender: TObject; aGene: TKnobsGene);
    Begin
      If Assigned( FOnWormChanged)
      Then FOnWormChanged( Self, aGene);
    End;

    Function    TKnobsWormPanel.GetWorm( anIndex: Integer): TKnobsWorm;
    Begin
      If ( anIndex >= 0) And ( anIndex< WormCount)
      Then Result := TKnobsWorm( FWorms[ anIndex])
      Else Result := Nil;
    End;

    Function    TKnobsWormPanel.GetData( wIndex, dIndex: Integer): TKnobsWormElt;
    Begin
      If ( wIndex>= 0) And ( wIndex< WormCount)
      Then Result := Worm[ wIndex][ dIndex]
      Else Result := 0;
    End;

    Function    TKnobsWormPanel.GetWormCount: Integer;
    Begin
      Result := FWorms.Count;
    End;

    Procedure   TKnobsWormPanel.SetRowCount( aValue : Integer);
    Begin
      If aValue <> FRowCount
      Then Begin
        FRowCount := aValue;
        UpdateCount;
      End;
    End;

    Procedure   TKnobsWormPanel.SetColCount( aValue : Integer);
    Begin
      If aValue <> FColCount
      Then Begin
        FColCount := aValue;
        UpdateCount;
      End;
    End;

    Procedure   TKnobsWormPanel.SetAudible( aValue : Integer);
    Begin
      If ( aValue >= 0) And ( aValue < WormCount)
      Then Worm[ aValue].Audible := True;
    End;

    Procedure   TKnobsWormPanel.SetFrozen( aValue : Integer);
    Begin
      If ( aValue >= 0) And ( aValue < WormCount)
      Then Worm[ aValue].Frozen := True;
    End;

    Procedure   TKnobsWormPanel.SetSpacing( aValue : Byte);
    Begin
      If aValue <> FSpacing
      Then Begin
        FSpacing := aValue;
        PlaceWorms;
      End;
    End;

    Procedure   TKnobsWormPanel.SetClearance( aValue : Byte);
    Begin
      If aValue <> FClearance
      Then Begin
        FClearance := aValue;
        UpdateWorms;
        PlaceWorms;
      End;
    End;

    Procedure   TKnobsWormPanel.SetWormWidth( aValue : Byte);
    Begin
      If aValue <> FWormWidth
      Then Begin
        FWormWidth := aValue;
        UpdateWorms;
        PlaceWorms;
      End;
    End;

    Procedure   TKnobsWormPanel.SetWormHeight( aValue : Byte);
    Begin
      If aValue <> FWormHeight
      Then Begin
        FWormHeight := aValue;
        UpdateWorms;
        PlaceWorms;
      End;
    End;

    Procedure   TKnobsWormPanel.SetWormSize( aValue : Integer);
    Var
      i : Integer;
    Begin
      If aValue <> FWormSize
      Then Begin
        FWormSize := aValue;
        For i := 0 To WormCount - 1 Do
          Worm[ i].Size := FWormSize;
      End;
    End;

    Procedure   TKnobsWormPanel.SetBackColor( aValue : TColor);
    Begin
      If aValue <> FBackColor
      Then Begin
        FBackColor := aValue;
        UpdateWorms;
      End;
    End;

    Procedure   TKnobsWormPanel.SetBorderColorNormal( aValue : TColor);
    Begin
      If aValue <> FBorderColorNormal
      Then Begin
        FBorderColorNormal := aValue;
        UpdateWorms;
      End;
    End;

    Procedure   TKnobsWormPanel.SetBorderColorAudible( aValue : TColor);
    Begin
      If aValue <> FBorderColorAudible
      Then Begin
        FBorderColorAudible := aValue;
        UpdateWorms;
      End;
    End;

    Procedure   TKnobsWormPanel.SetBorderColorFrozen( aValue : TColor);
    Begin
      If aValue <> FBorderColorFrozen
      Then Begin
        FBorderColorFrozen := aValue;
        UpdateWorms;
      End;
    End;

    Procedure   TKnobsWormPanel.SetBorderColorBoth( aValue : TColor);
    Begin
      If aValue <> FBorderColorBoth
      Then Begin
        FBorderColorBoth := aValue;
        UpdateWorms;
      End;
    End;

    Procedure   TKnobsWormPanel.SetPenColor( aValue : TColor);
    Begin
      If aValue <> FPenColor
      Then Begin
        FPenColor := aValue;
        UpdateWorms;
      End;
    End;

    Procedure   TKnobsWormPanel.SetTransparent( aValue : Boolean);
    Begin
      If aValue <> FTransparent
      Then Begin
        FTransparent := aValue;
        UpdateWorms;
      End;
    End;

    Procedure   TKnobsWormPanel.SetCrossOverProbability( aValue : TKnobsPercentage);
    Begin
      If aValue <> FCrossOverProbability
      Then Begin
        FCrossOverProbability := aValue;
        UpdateWorms;
      End;
    End;

    Procedure   TKnobsWormPanel.SetMutationProbability( aValue : TKnobsPercentage);
    Begin
      If aValue <> FMutationProbability
      Then Begin
        FMutationProbability := aValue;
        UpdateWorms;
      End;
    End;

    Procedure   TKnobsWormPanel.SetMutationRange( aValue : TKnobsWormElt);
    Begin
      If aValue <> FMutationRange
      Then Begin
        FMutationRange := aValue;
        UpdateWorms;
      End;
    End;

    Procedure   TKnobsWormPanel.SetMorphPercentage( aValue : TKnobsPercentage);
    Begin
      If aValue <> FMorphPercentage
      Then Begin
        FMorphPercentage := aValue;
        UpdateWorms;
      End;
    End;

    Procedure   TKnobsWormPanel.SetDrawLines( aValue : Boolean);
    Begin
      If aValue <> FDrawLines
      Then Begin
        FDrawLines := aValue;
        UpdateWorms;
      End;
    End;

// Protected

    Procedure   TKnobsWormPanel.Notification( aComponent: TComponent; anOperation: TOperation); // Override;
    Var
      i : Integer;
    Begin
      If aComponent Is TKnobsWorm
      Then Begin
        If anOperation = opInsert
        Then Begin
          FWorms.Add( aComponent);
          With TKnobsWorm( aComponent) Do
          Begin
            OnAudible := DoAudible;
            OnFrozen  := DoFrozen;
          End;
        End
        Else If anOperation = opRemove
        Then FWorms.Remove( aComponent);

        For i := 0 To WormCount - 1
        Do Worm[ i].Tag := i;
      End;
    End;

//  Public

    Constructor TKnobsWormPanel.Create( anOwner: TComponent); // Override;
    Begin
      Inherited Create( anOwner);
      FWorms               := TList.Create;
      OnResize             := DoResize;
      DoubleBuffered       := True;
      BevelInner           := bvLowered;
      BevelOuter           := bvRaised;
      Caption              := '.';
      Height               := 167;
      Width                := 167;
      Spacing              :=   2;
      Clearance            :=  10;
      WormWidth            :=  80;
      WormHeight           :=  80;
      WormSize             := 120;
      BackColor            := clWhite;
      BorderColorNormal    := clGray;
      BorderColorAudible   := clYellow;
      BorderColorFrozen    := clBlue;
      BorderColorBoth      := clLime;
      PenColor             := clBlack;
      Transparent          := False;
      CrossOverProbability :=  33;
      MutationProbability  :=  50;
      MutationRange        :=   5;
      MorphPercentage      :=  20;
      DrawLines            := True;
      Audible              :=  -1;
      Frozen               :=  -1;
      RowCount             :=   2;
      ColCount             :=   2;
    End;

    Destructor  TKnobsWormPanel.Destroy; // Override;
    Begin
      FWorms.Free;
      Inherited Destroy;
    End;

    Procedure   TKnobsWormPanel.FillRandom;
    Var
      i : Integer;
    Begin
      For i := 0 To WormCount - 1
      Do begin
        If i <> Frozen
        Then Worm[ i].FillRandom;
      end;
    End;

    Procedure   TKnobsWormPanel.Mutate;
    Var
      i : Integer;
    Begin
      For i := 0 To WormCount - 1
      Do begin
        If i <> Frozen
        Then Worm[ i].Mutate;
      end;
    End;

    Procedure   TKnobsWormPanel.Mate;
    Var
      i : Integer;
    Begin
      For i := 0 To WormCount - 1
      Do begin
        If ( i <> Audible) And ( i <> Frozen)
        Then Worm[ i].CopyFrom( Worm[ Audible]);
      end;
      For i := 0 To WormCount - 1
      Do begin
        If ( i <> Audible) And ( i <> Frozen)
        Then Worm[ i].MateWith( Worm[ Frozen]);
      end;
    End;

    Procedure   TKnobsWormPanel.Morph;
    Begin
      AudibleWorm.MorphTo( FrozenWorm);
    End;

    Procedure   TKnobsWormPanel.TriggerChange;
    Begin
      AudibleWorm.TriggerChange;
    End;

    Procedure   TKnobsWormPanel.SelectHPrev;
    Begin
      Audible := ( Audible + WormCount - 1) Mod WormCount;
    End;

    Procedure   TKnobsWormPanel.SelectHNext;
    Begin
      Audible := ( Audible + 1) Mod WormCount;
    End;

    Procedure   TKnobsWormPanel.SelectVPrev;
    Var
      aNext : Integer;
    Begin
      aNext := Audible - ColCount;
      If aNext = - ColCount
      Then aNext := WormCount - 1
      Else If aNext < 0
      Then aNext := aNext + WormCount - 1;
      Audible := aNext;
    End;

    Procedure   TKnobsWormPanel.SelectVNext;
    Var
      aNext : Integer;
    Begin
      aNext := Audible + ColCount;
      If aNext = WormCount + ColCount - 1
      Then aNext := 0
      Else If aNext >= WormCount
      Then aNext := aNext - WormCount + 1;
      Audible := aNext;
    End;

    Procedure   TKnobsWormPanel.ActivateHPrev;
    Begin
      Frozen := ( Frozen + WormCount - 1) Mod WormCount;
    End;

    Procedure   TKnobsWormPanel.ActivateHNext;
    Begin
      Frozen := ( Frozen + 1) Mod WormCount;
    End;

    Procedure   TKnobsWormPanel.ActivateVPrev;
    Var
      aNext : Integer;
    Begin
      aNext := Frozen - ColCount;
      If aNext = - ColCount
      Then aNext := WormCount - 1
      Else If aNext < 0
      Then aNext := aNext + WormCount - 1;
      Frozen := aNext;
    End;

    Procedure   TKnobsWormPanel.ActivateVNext;
    Var
      aNext : Integer;
    Begin
      aNext := Frozen + ColCount;
      If aNext = WormCount + ColCount - 1
      Then aNext := 0
      Else If aNext >= WormCount
      Then aNext := aNext - WormCount + 1;
      Frozen := aNext;
    End;


{========
  TGene = Class( TObject)
  Private
      FData  : PWormData;
      FSize  : Integer;
      FName  : String;
  Public
    Property    Size : Integer Read FSize Write SetSize;
  Private
}

    Procedure   TKnobsGene.SetSize( aValue: Integer);
    Begin
      If aValue <> FSize
      Then Begin
        If FSize > 0
        Then FreeMem( FData, FSize * SizeOf( TKnobsWormElt));
        FSize := aValue;
        If FSize > 0
        Then GetMem( FData, FSize * SizeOf( TKnobsWormElt));
      End;
    End;

    Function    TKnobsGene.GetData( anIndex: Integer): TKnobsWormElt;
    Begin
      Result := FData^[ anIndex];
    End;

    Procedure   TKnobsGene.SetData( anIndex: Integer; aValue: TKnobsWormElt);
    Begin
      FData^[ anIndex] := aValue;
    End;

//  Public

    Constructor TKnobsGene.Create( aWorm: TKnobsWorm);
    Begin
      Inherited Create;
      If Assigned( aWorm)
      Then Begin
        Size := aWorm.Size;
        Move( aWorm.FGene.FData^, FData^, Size * SizeOf( TKnobsWormElt));
      End;
    End;

    Destructor  TKnobsGene.Destroy;
    Begin
      Size := 0;
      Inherited Destroy;
    End;

    Procedure   TKnobsGene.Save( Var aFile: TextFile);
    Var
      i : Integer;
    Begin
      Write( aFile, Format( '%d', [ Size]));
      For i := 0 To Size - 1
      Do Write( aFile, Format( ', %d', [ Data[ i]]));
      WriteLn( aFile, ';');
    End;



{ ========
  TGeneBox = Class( TCustomListBox)
  Private
      FOnGetGeneName : TOnGetGeneName;
      FDragged       : TGene;
  Published
    Property    Align;
    Property    Anchors;
    Property    BiDiMode;
    Property    BorderStyle;
    Property    Color;
    Property    Columns;
    Property    Constraints;
    Property    Ctl3D;
    Property    DragCursor;
    Property    DragKind;
    Property    Enabled;
    Property    ExtendedSelect;
    Property    Font;
    Property    ImeMode;
    Property    ImeName;
    Property    IntegralHeight;
    Property    ItemHeight;
    Property    ParentBiDiMode;
    Property    ParentColor;
    Property    ParentCtl3D;
    Property    ParentFont;
    Property    ParentShowHint;
    Property    PopupMenu;
    Property    ShowHint;
    Property    Sorted;
    Property    Style;
    Property    TabOrder;
    Property    TabWidth;
    Property    Visible;
  Published
    Property    OnGetGeneName: TOnGetGeneName Read FOnGetGeneName Write FOnGetGeneName;
    Property    OnClick;
    Property    OnContextPopup;
    Property    OnDblClick;
    Property    OnEnter;
    Property    OnExit;
    Property    OnKeyDown;
    Property    OnKeyPress;
    Property    OnKeyUp;
    Property    OnMeasureItem;
  private
}

    Procedure   TKnobsGeneBox.Do_OnDragDrop( Sender, Source: TObject; X, Y: Integer);
    Begin
      If Source Is TKnobsWorm
      Then AddWorm( '', TKnobsWorm( Source));
    End;

    Procedure   TKnobsGeneBox.Do_OnDragOver( Sender, Source: TObject; X, Y: Integer; State: TDragState; Var Accept: Boolean);
    Begin
      Accept := ( Source Is TKnobsWorm) And ( Source <> Self);
    End;

    Procedure   TKnobsGeneBox.Do_OnEndDrag( Sender, Target: TObject; X, Y: Integer);
    Begin
    End;

    Procedure   TKnobsGeneBox.Do_OnMouseDown( Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
    Var
      i : Integer;
    Begin
      If Button = mbLeft
      Then begin
        With Sender as TKnobsGeneBox
        Do Begin
          i := ItemAtPos(Point( X, Y), True);
          If i >= 0
          Then Begin
            FDragged := TKnobsGene( Items.Objects[ i]);
            BeginDrag( False);
          End;
        end;
      end;
    End;

    Procedure   TKnobsGeneBox.AddWorm( aName: String; aWorm: TKnobsWorm);
    Var
      aGene : TKnobsGene;
    Begin
      aGene := TKnobsGene.Create( aWorm);
      If aName = ''
      Then Begin
        If Assigned( OnGetGeneName)
        Then OnGetGeneName( Self, aGene, aName);
      End;
      If aName <> ''
      Then Items.AddObject( aName, aGene);
    End;

//  Public

    Constructor TKnobsGeneBox.Create( anOwner: TComponent); // Override;
    Begin
      Inherited Create( anOwner);
      OnDragDrop  := Do_OnDragDrop;
      OnDragOver  := Do_OnDragOver;
      OnEndDrag   := Do_OnEndDrag;
      OnMouseDown := Do_OnMouseDown;
    End;

    Destructor  TKnobsGeneBox.Destroy; // Override;
    Begin
      If HasParent
      Then Clear;
      Inherited Destroy;
    End;

    Procedure   TKnobsGeneBox.Load( Const aFileName: String);
    Const
      Separators : TCharSet = [ ' ', ^I, ':', '=', ',', '.', ';'];
    Var
      aFile : TextFile;
      S     : String;
      Count : Integer;
      aName : String;
      i     : Integer;
      j     : Integer;
      p     : Integer;
      GCount: Integer;
      aGene : TKnobsGene;
    Begin
      AssignFile( aFile, aFileName);
      Reset( aFile);
      Try
        ReadLn( aFile, S);
        p     := 1;
        Count := ScanInt( S, p, Separators);
        For i := 1 To Count
        Do Begin
          ReadLn( aFile, S);
          p     := 1;
          aName := ScanName( S, p, Separators);
          If aName = ''
          Then Raise EKnobsInvalidFile.Create( 'Not a Friemel ?');
          GCount := ScanInt( S, p, Separators);
          aGene  := TKnobsGene.Create;
          With aGene
          Do Begin
            Size := GCount;
            For j := 0 To GCount - 1
            Do Data[ j] := ScanInt( S, p, Separators);
          End;
          Items.AddObject( aName, aGene);
        End;
      Finally
        CloseFile( aFile);
      End;
    End;

    Procedure   TKnobsGeneBox.Append( Const aFileName: String);
    Var
      aFile : TextFile;
      i     : Integer;
    Begin
      AssignFile( aFile, aFileName);
      If FileExists( aFileName)
      Then System.Append( aFile)
      Else Rewrite( aFile);
      Try
        With Items
        Do Begin
          WriteLn( aFile, Count);
          For i := 0 To Count - 1
          Do Begin
            Write( aFile, Format( '%s = ', [ Items[ i]]));
            TKnobsGene( Objects[ i]).Save( aFile);
          End;
        End;
      Finally
        CloseFile( aFile);
      End;
    End;

    Procedure   TKnobsGeneBox.Save( Const aFileName: String);
    Begin
      DeleteFile( aFileName);
      Append( aFileName);
    End;

    Procedure   TKnobsGeneBox.Clear;
    Begin
      With Items Do
        While Count > 0
        Do Begin
          Objects[ Count - 1].Free;
          Objects[ Count - 1] := Nil;
          Delete( Count - 1)
        End;
    End;

{ ========
  TSubstitution = Class
  Private
      FRule   : TStringList;
      FActive : PBoolArray;
      FCount  : Integer;
  Public
    Property    Active[ anIndex: Integer]: Boolean Read GetActive Write SetActive;
    Property    Rule  [ anIndex: Integer]: String  Read GetRule   Write SetRule;
    Property    Count                    : Integer Read FCount    Write SetCount;
  Private
}

    Procedure   TSubstitution.CheckBounds( anIndex: Integer);
    Begin
      If ( anIndex < 0) Or ( anIndex >= FCount)
      Then
        Raise
          EKnobsBounds.
            CreateFmt(
              'TSubstitution : Index ( %d) out of bounds [ %d, %d]',
              [
                anIndex,
                0,
                FCount - 1
              ]
            );
    End;

    Function    TSubstitution.GetActive( anIndex: Integer): Boolean;
    Begin
      CheckBounds( anIndex);
      Result := FActive^[ anIndex];
    End;

    Procedure   TSubstitution.SetActive( anIndex: Integer; aValue: Boolean);
    Begin
      CheckBounds( anIndex);
      FActive^[ anIndex] := aValue;
    End;

    Function    TSubstitution.GetRule( anIndex: Integer): String;
    Begin
      CheckBounds( anIndex);
      If anIndex < FRule.Count
      Then Result := FRule[ anIndex]
      Else Result := '';
    End;

    Procedure   TSubstitution.SetRule( anIndex: Integer; aValue: String);
    Begin
      CheckBounds( anIndex);
      While FRule.Count <= anIndex
      Do FRule.Add( '');
      FRule[ anIndex] := aValue;
    End;

    Procedure   TSubstitution.SetCount ( aValue: Integer);
    Var
      aBoolArray: PKnobsBoolArray;
    Begin
      If aValue <> FCount
      Then Begin
        aBoolArray := Nil;
        If aValue > 0
        Then Begin
          GetMem( aBoolArray, aValue * SizeOf( Boolean));
          FillChar( aBoolArray^, aValue * SizeOf( Boolean), 0);
        End;
        If ( aValue > 0) And ( FCount > 0)
        Then Move( FActive^, aBoolArray^, Min( aValue, FCount) * SizeOf( Boolean));
        If FCount > 0
        Then FreeMem( FActive, FCount * SizeOf( Boolean));
        FCount  := aValue;
        FActive := aBoolArray;
        While FRule.Count < Count
        Do FRule.Add( '');
      End;
    End;

//  Public

    Constructor TSubstitution.Create;
    Begin
      Inherited Create;
      FRule := TStringList.Create;
      FRule.Sorted := False;
    End;

    Destructor  TSubstitution.Destroy; // Override;
    Begin
      Count := 0;
      FRule.Free;
      Inherited Destroy;
    End;

    Procedure   TSubstitution.CopyFrom( aSubstitution: TSubstitution);
    Var
      i : Integer;
    Begin
      If Assigned( aSubstitution)
      Then Begin
        Count := aSubstitution.Count;
        FRule.Clear;
        For i := 0 To Count - 1
        Do Rule[ i] := aSubstitution.Rule[ i];
        If Count > 0
        Then Move( aSubstitution.FActive^, FActive^, Count * SizeOf( Boolean));
      End
      Else Count := 0;
    End;

    Procedure   TSubstitution.Load( Const AfileName: String);
    Const
      Separators : TCharSet = [ ' ', ^I];
    Var
      aFile  : TextFile;
      S      : String;
      p      : Integer;
      i      : Integer;
    Begin
      AssignFile( aFile, aFileName);
      Reset( aFile);
      Try
        ReadLn( aFile, S);
        p     := 1;
        Count := ScanInt( S, p, Separators);
        FRule.Clear;
        For i := 0 To Count - 1
        Do Begin
          ReadLn( aFile, S);
          FRule.Add( S);
        End;
        For i := 0 To Count - 1
        Do Begin
          ReadLn( aFile, S);
          p          := 1;
          Active[ i] := ScanBool( S, p, Separators);
        End;
      Finally
        CloseFile( aFile);
      End;
    End;


    Procedure   TSubstitution.Save( Const AfileName: String);
    Const
      Bools : Array[ Boolean] Of String = ( 'False', 'True');
    Var
      aFile : TextFile;
      i     : Integer;
    Begin
      AssignFile( aFile, aFileName);
      Rewrite( aFile);
      Try
        WriteLn( aFile, Count);
        For i := 0 To Count - 1
        Do WriteLn( aFile, Rule[ i]);
        For i:= 0 To Count - 1
        Do WriteLn( aFile, Bools[ Active[ i]]);
      Finally
        CloseFile( aFile);
      End;
    End;



Initialization

  RegisterClasses(
    [
      TKnobsWorm,
      TKnobsWormPanel
    ]
  );

Finalization

  UnregisterClasses(
    [
      TKnobsWorm,
      TKnobsWormPanel
    ]
  );


end.
