{
  polyphase filters, used for up and down-sampling
  original c++ code by Dave from Muon Software found
  at MusicDSP.

  Rewritten in pascal by Shannon Faulkner, 4/7/06.

  source: http://www.musicdsp.org/showone.php?id=223

  Changed a bit by Blue Hell for use in Wren with Delphi XE5
}

unit PolyPhaseFilter;

interface

type


  TPPAllPass = class
  private
    FA  : Double;
    FX0 : Double;
    FX1 : Double;
    FX2 : Double;
    FY0 : Double;
    FY1 : Double;
    FY2 : Double;
  public
    constructor Create( aCoefficient : Double);
    function    Process( anInput : Double) : Double;
  end;

  TPPDoubleArray  = array of Double;
  TPPAllPassArray = array of TPPAllPass;

  TPPAllPassFilterCascade = class
  private
    AllPassFilters : TPPAllPassArray;
    FOrder         : Integer;
  public
    constructor Create( aCoefficients : TPPDoubleArray);
    function    Process( anInput : Double) : Double;
  end;

  TPPHalfBandFilter = class
  private
    FOrder   : Integer;
    FACoeffs : TPPDoubleArray;
    FBCoeffs : TPPDoubleArray;
    FFilterA : TPPAllPassFilterCascade;
    FFilterB : TPPAllPassFilterCascade;
    FOldOut  : Double;
  public
    constructor Create( anOrder : Integer; IsSteep : Boolean);
    function    Process( anInput : Double) : Double;
  end;



implementation



{ ========
  TPPAllPass = class
  private
    FA  : Double;
    FX0 : Double;
    FX1 : Double;
    FX2 : Double;
    FY0 : Double;
    FY1 : Double;
    FY2 : Double;
  public
}

    constructor TPPAllPass.Create( aCoefficient : Double);
    begin
      FA  := aCoefficient;

      FX0 := 0;
      FX1 := 0;
      FX2 := 0;

      FY0 := 0;
      FY1 := 0;
      FY2 := 0;
    end;

    function    TPPAllPass.Process( anInput : Double) : Double;
    begin
      FX2 := FX1;                                  // shuffle inputs
      FX1 := FX0;
      FX0 := anInput;

      FY2 := FY1;                                  // shuffle outputs
      FY1 := FY0;

      Result := FX2 + (( anInput - FY2) * FA);     // allpass filter 1
      FY0    := Result;
    end;


{ ========
  TPPAllPassFilterCascade = class
  private
    AllPassFilters : TAllpassArray;
    FOrder         : Integer;
  public
}

    constructor TPPAllPassFilterCascade.Create( aCoefficients : TPPDoubleArray);
    var
      i : Integer;
    begin
      FOrder := Length( aCoefficients);
      SetLength( AllPassFilters, FOrder);

      for i := 0 to FOrder - 1
      do AllPassFilters[ i] := TPPAllPass.Create( aCoefficients[ i]);
    end;

    function    TPPAllPassFilterCascade.Process( anInput : Double) : Double;
    var
      i : Integer;
    begin
      Result := anInput;

      for i := 0 to FOrder - 1
      do Result := AllPassFilters[ i].Process( Result);
    end;


{ ========
  TPPHalfBandFilter = class
  private
    FOrder   : Integer;
    FACoeffs : TDoubleArray;
    FBCoeffs : TDoubleArray;
    FFilterA : TAllPassFilterCascade;
    FFilterB : TAllPassFilterCascade;
    FOldOut  : Double;
  public
}

    constructor TPPHalfBandFilter.Create( anOrder : Integer; IsSteep : Boolean);
    begin
      Assert( anOrder in [ 2, 4, 6, 8, 10, 12]);

      FOrder := anOrder;

      if IsSteep
      then begin
        if anOrder = 12 // rejection = 104 dB, transition band = 0.01
        then begin
          FACoeffs := TPPDoubleArray.Create( 0.036681502163648017, 0.2746317593794541 , 0.56109896978791948, 0.769741833862266, 0.8922608180038789, 0.962094548378084 );
          FBCoeffs := TPPDoubleArray.Create( 0.13654762463195771 , 0.42313861743656667, 0.6775400499741616 , 0.839889624849638, 0.9315419599631839, 0.9878163707328971);
        end
        else if anOrder = 10 // rejection = 86 dB, transition band = 0.01
        then begin
          FACoeffs := TPPDoubleArray.Create( 0.051457617441190984, 0.35978656070567017, 0.6725475931034693, 0.8590884928249939, 0.9540209867860787);
          FBCoeffs := TPPDoubleArray.Create( 0.18621906251989334 , 0.529951372847964  , 0.7810257527489514, 0.9141815687605308, 0.985475023014907 );
        end
        else if anOrder = 8  // rejection = 69 dB, transition band = 0.01
        then begin
          FACoeffs := TPPDoubleArray.Create( 0.07711507983241622, 0.4820706250610472, 0.7968204713315797, 0.9412514277740471);
          FBCoeffs := TPPDoubleArray.Create( 0.2659685265210946 , 0.6651041532634957, 0.8841015085506159, 0.9820054141886075);
        end
        else if anOrder = 6  // rejection = 51 dB, transition band = 0.01
        then begin
          FACoeffs := TPPDoubleArray.Create( 0.1271414136264853 , 0.6528245886369117, 0.9176942834328115);
          FBCoeffs := TPPDoubleArray.Create( 0.40056789819445626, 0.8204163891923343, 0.9763114515836773);
        end
        else if anOrder = 4  // rejection = 53 dB, transition band = 0.05
        then begin
          FACoeffs := TPPDoubleArray.Create( 0.12073211751675449, 0.6632020224193995);
          FBCoeffs := TPPDoubleArray.Create( 0.3903621872345006 , 0.890786832653497 );
        end
        else begin           // order = 2, rejection = 36 dB, transition band = 0.1
          FACoeffs := TPPDoubleArray.Create( 0.23647102099689224);
          FBCoeffs := TPPDoubleArray.Create( 0.7145421497126001 );
        end;
      end

      else begin             // not steep, softer slopes, more attenuation and less stopband ripple
        if anOrder = 12      // rejection = 104 dB, transition band = 0.01
        then
        begin
          FACoeffs := TPPDoubleArray.Create( 0.01677466677723562, 0.13902148819717805, 0.3325011117394731, 0.53766105314488  , 0.7214184024215805 , 0.8821858402078155);
          FBCoeffs := TPPDoubleArray.Create( 0.06501319274445962, 0.23094129990840923, 0.4364942348420355, 0.6329609551399348, 0.80378086794111226, 0.9599687404800694);
        end
        else if anOrder = 10 // rejection = 86 dB, transition band = 0.01
        then begin
          FACoeffs := TPPDoubleArray.Create( 0.02366831419883467, 0.18989476227180174, 0.43157318062118555, 0.6632020224193995, 0.860015542499582  );
          FBCoeffs := TPPDoubleArray.Create( 0.09056555904993387, 0.3078575723749043 , 0.5516782402507934 , 0.7652146863779808, 0.95247728378667541);
        end
        else if anOrder = 8  // rejection = 69 dB, transition band = 0.01
        then begin
          FACoeffs := TPPDoubleArray.Create( 0.03583278843106211, 0.2720401433964576, 0.5720571972357003, 0.827124761997324 );
          FBCoeffs := TPPDoubleArray.Create( 0.1340901419430669 , 0.4243248712718685, 0.7062921421386394, 0.9415030941737551);
        end
        else if anOrder = 6  // rejection = 51 dB, transition band=0.01
        then begin
          FACoeffs := TPPDoubleArray.Create( 0.06029739095712437, 0.4125907203610563, 0.7727156537429234);
          FBCoeffs := TPPDoubleArray.Create( 0.21597144456092948, 0.6043586264658363, 0.9238861386532906);
        end
        else if anOrder = 4  // rejection = 53 dB,transition band = 0.05
        then begin
          FACoeffs := TPPDoubleArray.Create( 0.07986642623635751, 0.5453536510711322);
          FBCoeffs := TPPDoubleArray.Create( 0.28382934487410993, 0.8344118914807379);
        end
        else begin           // order = 2, rejection = 36 dB, transition band = 0.1
          FACoeffs := TPPDoubleArray.Create( 0.23647102099689224);
          FBCoeffs := TPPDoubleArray.Create( 0.7145421497126001 );
        end;
      end;

      FFilterA := TPPAllPassFilterCascade.Create( FACoeffs);
      FFilterB := TPPAllPassFilterCascade.Create( FBCoeffs);

      FOldOut := 0;
    end;

    function    TPPHalfBandFilter.Process( anInput : Double) : Double;
    begin
      Result  := 0.5 * ( FFilterA.Process( anInput) + FOldOut);
      FOldOut := FFilterB.Process( anInput);
    end;

end.

