unit SVFUnit;

// Source : http://www.musicdsp.org/showone.php?id=142

interface

type

  TSvfChamberlin = class
  private
    FInternalQ    : Double;
    FQ            : Double;
    FInternalFreq : Double;
    FFrequency    : Double;
    FSampleRate   : Double;
    FReciprocalSR : Double;
    FDelay1       : Double;
    FDelay2       : Double;
  private
    procedure   SetFrequency ( const aValue : Double);
    procedure   SetSampleRate( const aValue : Double);
    procedure   SetQ         ( const aValue : Double);
  public
    constructor Create;
    procedure   Process( const anInput : Double; var aLowPass, aBandPass, aBandReject, aHighPass : Double);
  public
    property    Frequency  : Double read FFrequency  write SetFrequency;
    property    SampleRate : Double read FSampleRate write SetSampleRate;
    property    Q          : Double read FQ          write SetQ;
  end;



implementation


uses

  sysutils;

const

  K_DENORM     = 1.0e-24;
  MIN_SAMPRATE = 44100.0;                 // Minimal sample rate, in Hz


    constructor TSvfChamberlin.Create;
    begin
      inherited;
      SampleRate := MIN_SAMPRATE;         // First set sample rate, then Frequency
      Frequency  := 329.6;                // in Hz
      Q          :=   1.0;                // Q can be set at arbitrary time
    end;

    procedure   TSvfChamberlin.SetFrequency( const aValue : Double);
    var
      aFreq : Double;
    begin
      if aValue > FSampleRate / 4
      then aFreq := FSampleRate / 4
      else aFreq := aValue;

      if aFreq <> FFrequency
      then begin
        FFrequency    := aFreq;
        FInternalFreq := FFrequency * FReciprocalSR; // This is inexact for speed reasons
      end;
    end;

    procedure   TSvfChamberlin.SetSampleRate( const aValue : Double);
    var
      aRate : Double;
    begin
      if aValue <= MIN_SAMPRATE
      then aRate := MIN_SAMPRATE
      else aRate := aValue;

      if aRate <> FSampleRate
      then begin
        FSampleRate   := aValue;
        FReciprocalSR := 2 * Pi / FSampleRate;
      end;
    end;

    procedure   TSvfChamberlin.SetQ( const aValue : Double);
    begin
      if aValue <> FQ
      then begin
        if aValue > 0.5
        then FQ := aValue
        else FQ := 0.5;

        FInternalQ := 1 / FQ;
      end;
    end;

    procedure   TSvfChamberlin.Process( const anInput : Double; var aLowPass, aBandPass, aBandReject, aHighPass : Double);
    begin
      aLowPass    := FDelay2 + FInternalFreq * FDelay1 - K_DENORM;
      aHighPass   := anInput - aLowPass - FInternalQ * FDelay1;
      aBandPass   := FInternalFreq * aHighPass + FDelay1;
      aBandReject := aHighPass + aLowPass;
      FDelay1     := aBandPass;
      FDelay2     := K_DENORM + aLowPass;
    end;

end.

